#ifdef CoCoA_WITH_NORMALIZ
//   Copyright (c)  2010-2012 Anna M. Bigatti, Christof Soeger

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for Normaliz integration

#include "CoCoA/ExternalLibs-Normaliz.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/IntOperations.H"
#include "CoCoA/TmpPPVector.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"

#include "CoCoA/io.H"  // just for debugging
#include <iostream>  // just for debugging
#include <sstream>  // for ErrorMessage

#include "gmpxx.h"
#include "libnormaliz/cone.h"
using libnormaliz::Cone;
using libnormaliz::ConeProperties;



// #include <list>
// using std::list;
#include <map>
using std::map;
#include <vector>
using std::vector;



namespace CoCoA
{


  class TemporarilySet
  {
  public:
    TemporarilySet(bool& var, bool value): myVar(var), myOrigValue(var) { myVar = value; }
    ~TemporarilySet() { myVar = myOrigValue; }
  private: // data members
    bool& myVar;
    bool myOrigValue;
  };

  namespace Normaliz
  {

    libnormaliz::InputType ToInputType(const std::string& TypeString)
    {
      return libnormaliz::to_type(TypeString);
    }

    class ConeImpl: protected IntrusiveReferenceCount
    {
      friend class SmartPtrIRC<const ConeImpl>; // Morally "friend Cone", so it can alter reference count.

      public:
        ConeImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
        ConeImpl(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m);
        ConeImpl(const PPVector& v, libnormaliz::Type::InputType InputType);
        friend std::vector<std::vector<BigInt> > HilbertBasis(const cone& c);
        friend std::vector<std::vector<BigInt> > Deg1Elements(const cone& c);
        friend std::vector<std::vector<BigInt> > SupportHyperplanes(const cone& c);
        friend std::vector<BigInt> HVector(const cone& c);

      private:
        void myComputation(const libnormaliz::ConeProperties& CPs) const;
        void myComputation(libnormaliz::ConeProperty::Enum CP) const;    

      private:
        mutable bool mySmallConeIsGood;
        mutable libnormaliz::Cone<long> myConeLong; // be careful!
        mutable libnormaliz::Cone<mpz_class> myConeMPZ; // be careful!
    };


    // implementation of cone

    cone::cone(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType) : mySmartPtr(new ConeImpl(v, InputType)) {}

    cone::cone(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m) : mySmartPtr(new ConeImpl(m)) {}

    cone::cone(ConstMatrixView M, libnormaliz::Type::InputType InputType) : mySmartPtr(new ConeImpl(MatrixToVecVecBigInt(M), InputType)) {}

    cone::cone(const ConeImpl* ptr): mySmartPtr(ptr) {}
    cone::cone(const cone& c) : mySmartPtr(c.mySmartPtr) {}
    cone::~cone() {}
    const ConeImpl* cone::operator->() const { return mySmartPtr.operator->(); }

    namespace  // conversion functions, implementation at the end of the file
    {
      std::vector<BigInt> LongVToBigIntV(const std::vector<long>& VIn);
      std::vector<BigInt> MPZ_classVToBigIntV(const std::vector<mpz_class>& VIn);
      std::vector<long> BigIntVToLongV(const std::vector<BigInt>& VIn);
      std::vector<mpz_class> BigIntVToMPZ_classV(const std::vector<BigInt>& VIn);
      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<long> >& l);
      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<mpz_class> >& l);
      std::vector<std::vector<long> >  ReturnVecVecLong(const std::vector<std::vector<BigInt> >& v);
      map< libnormaliz::InputType, vector<vector<long> > >  ReturnMapVecVecLong(const map< libnormaliz::InputType, vector<vector<BigInt> > > m);
      std::vector<std::vector<mpz_class> >  ReturnVecVecMPZ_class(const std::vector<std::vector<BigInt> >& v);
      map< libnormaliz::InputType, vector<vector<mpz_class> > >  ReturnMapVecVecMPZ_class(const map< libnormaliz::InputType, vector<vector<BigInt> > > m);
      std::vector<std::vector<BigInt> > PPVectorToVecVecBigInt(const PPVector& ppv);
      void VecVecBigIntToPPVector(PPVector& ppv, const std::vector<std::vector<BigInt> >& M);
    } // end of anonymous namespace

    //this is used in CoCoA5 methods, so not in the anonymous namespace
    std::vector<std::vector<BigInt> > MatrixToVecVecBigInt(ConstMatrixView M);


    // implementation of ConeImpl
    ConeImpl::ConeImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType):
          myConeLong(ReturnVecVecLong(v), InputType),
          myConeMPZ(ReturnVecVecMPZ_class(v), InputType)
      {
        mySmallConeIsGood = (myConeLong.getBasisChange().get_rank() > 0);
      }

    ConeImpl::ConeImpl(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m):
      myConeLong(ReturnMapVecVecLong(m)),
      myConeMPZ(ReturnMapVecVecMPZ_class(m))
    {}

    void ConeImpl::myComputation(const libnormaliz::ConeProperties& CPs) const
    {
      try
      {
        if (mySmallConeIsGood)
        {
          TemporarilySet save(libnormaliz::test_arithmetic_overflow, true);
          libnormaliz::ConeProperties missing = myConeLong.compute(CPs);
          if (missing.any())
          {
            std::ostringstream os;
            os << missing;
            CoCoA_ERROR("Normaliz cannot compute "+ os.str(), "myComputation long");
            return;
          }
        }
      }
      catch (libnormaliz::ArithmeticException&) { mySmallConeIsGood = false; }
      myConeMPZ.compute(CPs);
      libnormaliz::ConeProperties missing = myConeMPZ.compute(CPs);
      if (missing.any())
      {
        std::ostringstream os;
        os << missing;
        CoCoA_ERROR("Normaliz cannot compute "+ os.str(), "myComputation long");
      }
    }

    void ConeImpl::myComputation(libnormaliz::ConeProperty::Enum CP) const
    {
      myComputation(ConeProperties(CP));
    }

    // friend functions which are the interface to the user
    std::vector<std::vector<BigInt> >  HilbertBasis(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::HilbertBasis);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getHilbertBasis());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getHilbertBasis());
      return v;
    }

    std::vector<std::vector<BigInt> > Deg1Elements(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::Deg1Elements);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getDeg1Elements());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getDeg1Elements());
      return v;
    }

    std::vector<std::vector<BigInt> > SupportHyperplanes(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getSupportHyperplanes());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getSupportHyperplanes());
      return v;
    }

    std::vector<BigInt> HVector(const cone& c)
    { 
      c->myComputation(libnormaliz::ConeProperty::HilbertSeries);
      if (c->mySmallConeIsGood)
        return MPZ_classVToBigIntV(c->myConeLong.getHilbertSeries().getNum());
      else
        return MPZ_classVToBigIntV(c->myConeMPZ.getHilbertSeries().getNum());
    }


    // applications to monomials
    PPVector NormalToricRing(const PPVector& ppv)
    { 
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::normalization);
      std::vector<std::vector<BigInt> > res = HilbertBasis(c);
      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res,res);
      return ppv_res;
    }

    PPVector IntClosureToricRing(const PPVector& ppv)
    {
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::integral_closure);
      std::vector<std::vector<BigInt> > res = HilbertBasis(c);
      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res,res);
      return ppv_res;
    }

    PPVector IntClosureMonIdeal(const PPVector& ppv)
    {
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::rees_algebra);
      vector<vector<BigInt> > res = HilbertBasis(c);
      vector<vector<BigInt> > res1;
      for (vector<vector<BigInt> >::const_iterator it=res.begin(); it!=res.end(); ++it) {
        BigInt tmp = it->back();
        if (it->back() == 1 ) {
          res1.push_back(*it);
          res1.back().pop_back();
        }
      }

      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res1(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res1,res1);
      return ppv_res1;
    }
    
    //if you want also the whole normalization of the rees algebra you have to specify an extra symbol
    //it may be done by NewSymbols(1), which creates an unique new symbol (with funny name)
/*    pair<PPVector,PPVector> IntegralClosureMonomialIdeal(const PPVector& ppv, symbol var)
    { 
      //check if the symbol is in the ring
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::rees_algebra);
      vector<vector<BigInt> > res = HilbertBasis(c);
      vector<vector<BigInt> > res1;
      for (vector<vector<BigInt> >::const_iterator it=res.begin(); it!=res.end(); ++it) {
        BigInt tmp = it->back();
        if (it->back() == 1 ) {
          res1.push_back(*it);
          res1.back().pop_back();
        }
      }
      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res1(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res1,res1);
      // create new PPM with extra symbol
      PPVector ppv_res(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res,res1);

      return make_pair(ppv_res1, ppv_res);
    }
*/     
    //We also want to return the  Here we have to change the PPMonoid by extending it or remove an unused var from the input

/*    PPVector EhrhartRing(const PPVector& ppv)
    {
    }
*/


/********************************************************
 ***               conversion functions               ***
 ********************************************************/

    namespace  // conversion functions
    {
      std::vector<BigInt> LongVToBigIntV(const std::vector<long>& VIn)
      {
        std::vector<BigInt> v;
        v.reserve(VIn.size());
        for (vector<long>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigInt(*it));
        return v;
      }


      std::vector<BigInt> MPZ_classVToBigIntV(const std::vector<mpz_class>& VIn)
      {
        std::vector<BigInt> v;
        v.reserve(VIn.size());
        for (vector<mpz_class>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigInt(it->get_mpz_t()));
        return v;
      }


      std::vector<long> BigIntVToLongV(const std::vector<BigInt>& VIn)
      {
        std::vector<long> v;
        long n;
        v.reserve(VIn.size());
        for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          if (IsConvertible(n, *it))  v.push_back(n);
          else CoCoA_ERROR(ERR::BadConvert, "normaliz: BigIntVToLongV");
        return v;
      }


      std::vector<mpz_class> BigIntVToMPZ_classV(const std::vector<BigInt>& VIn)
      {
        std::vector<mpz_class> v;
        v.reserve(VIn.size());
        for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(mpz_class(mpzref(*it)));
        return v;
      }


      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<long> >& l)
      {
        v.clear();      // not exception safe
        v.reserve(l.size());
        //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
        for (vector<vector<long> >::const_iterator it=l.begin(); it!=l.end(); ++it)
          v.push_back(LongVToBigIntV(*it));
      }    


      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<mpz_class> >& l)
      {
        v.clear();      // not exception safe
        v.reserve(l.size());
        //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
        for (vector<vector<mpz_class> >::const_iterator it=l.begin(); it!=l.end(); ++it)
          v.push_back(MPZ_classVToBigIntV(*it));
      }    


      std::vector<std::vector<long> >  ReturnVecVecLong(const std::vector<std::vector<BigInt> >& v)
      {
        std::vector<std::vector<long> > l;
        //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
        try{
            for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
              l.push_back(BigIntVToLongV(*it));
            //      std::clog << l << std::endl;
        } catch(...) {l.clear();}
        return l;
      }


      map< libnormaliz::InputType, vector<vector<long> > >  ReturnMapVecVecLong(const map< libnormaliz::InputType, vector<vector<BigInt> > > m)
      {
        map< libnormaliz::InputType, vector<vector<long> > > ret;
        for (map<libnormaliz::InputType, vector<vector<BigInt> > >::const_iterator it=m.begin(); it!=m.end(); ++it)
          ret.insert(make_pair((*it).first, ReturnVecVecLong((*it).second)));
        return ret;
      }
      
      
      std::vector<std::vector<mpz_class> >  ReturnVecVecMPZ_class(const std::vector<std::vector<BigInt> >& v)
      {
        std::vector<std::vector<mpz_class> > l;
        //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
        for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
          l.push_back(BigIntVToMPZ_classV(*it));
        //      std::clog << l << std::endl;
        return l;
      }


      map< libnormaliz::InputType, vector<vector<mpz_class> > >  ReturnMapVecVecMPZ_class(const map< libnormaliz::InputType, vector<vector<BigInt> > > m)
      {
        map< libnormaliz::InputType, vector<vector<mpz_class> > > ret;
        for (map<libnormaliz::InputType, vector<vector<BigInt> > >::const_iterator it=m.begin(); it!=m.end(); ++it)
          ret.insert(make_pair((*it).first, ReturnVecVecMPZ_class((*it).second)));
        return ret;
      }


      std::vector<std::vector<BigInt> > PPVectorToVecVecBigInt(const PPVector& ppv)
      {
        vector<BigInt> tmp;
        const long n =  len(ppv);
        vector<vector<BigInt> > v(n);
        for (long i=0; i<n; ++i)
        {
          BigExponents(tmp,PP(ppv[i]));
          v[i]=tmp;
        }
        return v;
      }

      void VecVecBigIntToPPVector(PPVector& ppv, const std::vector<std::vector<BigInt> >& M)
      {
        PPMonoid ppm = PPM(ppv);
        const long n =  len(M);
        for (long i=0; i < n; ++i)
        {
          ppv.myPushBack(PPMonoidElem(ppm, M[i]));
        }
      }

    } // end of anonymous namespace

    std::vector<std::vector<BigInt> > MatrixToVecVecBigInt(ConstMatrixView M)
    {
      BigInt tmp;
      vector<vector<BigInt> > v;
      for (long i=0; i<NumRows(M); ++i)
      {
        v.push_back(vector<BigInt>());
        for (long j=0; j<NumCols(M); ++j)
        {
          if (!IsInteger(tmp, M(i,j)))
            CoCoA_ERROR("Not an integer", "MatrixToVecVecBigInt");
          v[i].push_back(tmp);
        }
      }
      return v;
    }





  } // namespace Normaliz
} // namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ExternalLibs-Normaliz.C,v 1.17 2013/03/15 17:49:17 abbott Exp $
// $Log: ExternalLibs-Normaliz.C,v $
// Revision 1.17  2013/03/15 17:49:17  abbott
// Minor cleaning; removed an out-of-date TODO comment.
//
// Revision 1.16  2012/10/08 13:52:02  bigatti
// -- more cleaning and updates by Christof Soeger
//
// Revision 1.15  2012/10/05 10:17:04  bigatti
// by Christof Soeger
// * Made the NewCone functions to constructors of cone.
// * Introduced some abbreviation in the method names:
//  IntegralClosure -> IntClosure
//  MonomialIdeal -> MonIdeal
//  Normaliz -> Nmz  (the prefix for CoCoA5 functions)
// * New function IntClosureMonIdeal
//
// Revision 1.14  2012/09/28 13:56:36  abbott
// Cleaned up code with Christof's help -- now cleaner & more symmetrical.
// Also fixed a subtle bug where mySmallConeIsGood was not updated as it should have been.
//
// Revision 1.13  2012/08/03 16:31:22  bigatti
// -- changed: procedural --> functional (by C.Soeger)
//
// Revision 1.12  2012/07/25 12:46:27  bigatti
// -- merged with C.Soeger changes for NormalizComputation (init from map, etc)
//
// Revision 1.11  2012/07/19 17:12:05  abbott
// Added NewCone -- unified pseudo-ctor so user does not have to choose between long and BigInt.
//
// Revision 1.10  2012/06/19 14:44:41  bigatti
// -- changed Ht1 --> Deg1, changed def of HVector (by C.Soeger)
//
// Revision 1.9  2011/11/07 11:09:32  bigatti
// -- new ctors taking ConstRefRingElem
// -- added PPVector NormalToricRing(const PPVector& ppv)
// -- removed void HilbertBasis(std::vector<std::vector<BigInt> >& v, const cone& c)
//
// Revision 1.8  2011/10/04 13:03:02  bigatti
// -- new logo for gui
//
// Revision 1.7  2011/09/30 12:55:44  bigatti
// -- introduced namespace "Normaliz" and removed Normaliz from function names
// -- input of Normaliz functions in CoCoA-5 is now a matrix instead of
//    vector<vector<BigInt>>
//
// Revision 1.6  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.5  2011/07/20 15:31:21  bigatti
// -- fixed InputType in (undefined) functions
//
// Revision 1.4  2011/07/20 13:49:37  bigatti
// -- added "Normaliz" postfix to Normaliz function calls
//
// Revision 1.3  2011/07/20 12:45:12  bigatti
// -- new normaliz interface (not yet public)
//
// Revision 1.2  2011/02/17 16:50:04  bigatti
// -- getting ready for new official veson on Normaliz: added HVector, removed Triangulation
//
// Revision 1.1  2010/11/05 14:27:18  bigatti
// -- was TmpNormaliz**
//
// Revision 1.6  2010/10/12 11:22:54  bigatti
// -- TmpNormaliz.H simplified:
//    now cone is a smart pointer to ConeBase
//    and the concrete classes are entirely in the .C file
// -- added Ht1Elements, SupportHyperplanes, Triangulation
// -- added some text in ex-Normaliz1 and 2
//
// Revision 1.5  2010/10/08 13:40:37  bigatti
// -- moved inclusion of libnormaliz.cpp and instantiation of types
//    into TmpNormalizTypes.C
//
// Revision 1.4  2010/10/08 10:39:32  bigatti
// -- extended interface for normaliz
//

#endif
